package crypto

import (
	"errors"
	"fmt"

	"github.com/alexandrevicenzi/unchained"
)

var (
	ErrHashPassword    = errors.New("crypto: hashing password")
	ErrInvalidPassword = errors.New("crypto: password verification error")
	ErrWeakHasher      = errors.New("crypto: weak hasher")
)

const (
	DefaultHasher   = unchained.DefaultHasher
	DefaultSaltSize = unchained.DefaultSaltSize
)

type Hasher interface {
	// Hash hashes the specified password using a configured hasher algorithm.
	Hash(pw string) (string, error)
	// Verify verifies the specified hash and password
	Verify(pw, hash string) (bool, error)
	// Identify returns the name of the hasher implementation that produced the specified hashed string
	Identify(hash string) string
}

func NewHasher(cfg HasherConfig) (Hasher, error) {
	if cfg.Name == "" {
		cfg.Name = DefaultHasher
	}
	if cfg.SaltSize == 0 {
		cfg.SaltSize = DefaultSaltSize
	}

	// Lets not allow weak hashing implementations
	if unchained.IsWeakHasher(cfg.Name) {
		return nil, fmt.Errorf("%s is invalid, %w", cfg.Name, ErrWeakHasher)
	}

	return &unchainedHasher{
		hasher:   cfg.Name,
		saltSize: cfg.SaltSize,
	}, nil
}

// unchainedHasher implements the Hasher interface delegating hashing and verification into
// the implementations that are provided by unchained.
type unchainedHasher struct {
	hasher   string
	saltSize int
}

func (u *unchainedHasher) Hash(pw string) (string, error) {
	hash, err := unchained.MakePassword(pw, unchained.GetRandomString(u.saltSize), u.hasher)
	if err != nil {
		return "", fmt.Errorf("%v, %w", err, ErrHashPassword)
	}

	return hash, nil
}

func (u *unchainedHasher) Verify(pw, hash string) (bool, error) {
	isValid, err := unchained.CheckPassword(pw, hash)
	if err != nil {
		return false, fmt.Errorf("%v, %w", err, ErrInvalidPassword)
	}

	return isValid, nil
}

func (u *unchainedHasher) Identify(hash string) string {
	return unchained.IdentifyHasher(hash)
}
