package trace

import (
	"context"
	"io"

	"github.com/opentracing/opentracing-go"
	jaegercfg "github.com/uber/jaeger-client-go/config"

	"github.com/greenqloud/heimdall/pkg/logging"
)

type Config struct {
	ServiceName string
}

type ContextTracer interface {
	SpanFromContext(ctx context.Context) opentracing.Span
	StartSpanFromContext(ctx context.Context, operationName string, opts ...opentracing.StartSpanOption) (opentracing.Span, context.Context)
}

type Tracer interface {
	ContextTracer
	opentracing.Tracer
}

func NoOpTracer() Tracer {
	return &tracer{Tracer: opentracing.NoopTracer{}, logger: logging.NoOp()}
}

type tracer struct {
	opentracing.Tracer
	logger *logging.Logger
}

func NewTracer(cfg Config, logger *logging.Logger) (Tracer, io.Closer, error) {
	config := &jaegercfg.Configuration{
		ServiceName: cfg.ServiceName,
		Sampler: &jaegercfg.SamplerConfig{
			Type:  "const",
			Param: 1,
		},
		Reporter: &jaegercfg.ReporterConfig{
			LogSpans: true,
		},
	}

	t, closer, err := config.NewTracer()
	if err != nil {
		return nil, nil, err
	}
	return &tracer{Tracer: t, logger: logger}, closer, nil
}

func (t *tracer) StartSpan(operationName string, opts ...opentracing.StartSpanOption) opentracing.Span {
	t.logger.Debugw("Starting span", "operation", operationName)
	return &span{
		Span:   t.Tracer.StartSpan(operationName, opts...),
		logger: t.logger.With("span", operationName),
	}
}

func (t *tracer) Inject(sm opentracing.SpanContext, format interface{}, carrier interface{}) error {
	return t.Tracer.Inject(sm, format, carrier)
}

func (t *tracer) Extract(format interface{}, carrier interface{}) (opentracing.SpanContext, error) {
	return t.Tracer.Extract(format, carrier)
}

func (t *tracer) StartSpanFromContext(ctx context.Context, operationName string, opts ...opentracing.StartSpanOption) (opentracing.Span, context.Context) {
	t.logger.Debugw("Starting span from context", "operation", operationName)
	if s := t.SpanFromContext(ctx); s != nil {
		t.logger.Debugw("Span exists creating child span")
		opts = append(opts, opentracing.ChildOf(s.Context()))
	}

	s := t.StartSpan(operationName, opts...)
	return s, opentracing.ContextWithSpan(ctx, s)
}

func (t *tracer) SpanFromContext(ctx context.Context) opentracing.Span {
	t.logger.Debugw("Getting span from context")
	return opentracing.SpanFromContext(ctx)
}
